// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "simd/simd.h"

namespace starrocks::vectorized {
template <PrimitiveType PT>
const Buffer<typename JoinBuildFunc<PT>::CppType>& JoinBuildFunc<PT>::get_key_data(
        const JoinHashTableItems& table_items) {
    if (table_items.key_columns[0]->is_nullable()) {
        auto* null_column = ColumnHelper::as_raw_column<NullableColumn>(table_items.key_columns[0]);
        return ColumnHelper::as_raw_column<ColumnType>(null_column->data_column())->get_data();
    }

    return ColumnHelper::as_raw_column<ColumnType>(table_items.key_columns[0])->get_data();
}

template <PrimitiveType PT>
Status JoinBuildFunc<PT>::construct_hash_table(JoinHashTableItems* table_items,
                                               HashTableProbeState* probe_state) {
    auto& data = get_key_data(*table_items);
    if (table_items->key_columns[0]->is_nullable()) {
        auto* nullable_column =
                ColumnHelper::as_raw_column<NullableColumn>(table_items->key_columns[0]);
        auto& null_array = nullable_column->null_column()->get_data();
        for (size_t i = 1; i < table_items->row_count + 1; i++) {
            if (null_array[i] == 0) {
                uint32_t bucket_num = JoinHashMapHelper::calc_bucket_num<CppType>(
                        data[i], table_items->bucket_size);
                table_items->next[i] = table_items->first[bucket_num];
                table_items->first[bucket_num] = i;
            }
        }
    } else {
        for (size_t i = 1; i < table_items->row_count + 1; i++) {
            uint32_t bucket_num =
                    JoinHashMapHelper::calc_bucket_num<CppType>(data[i], table_items->bucket_size);
            table_items->next[i] = table_items->first[bucket_num];
            table_items->first[bucket_num] = i;
        }
    }
    return Status::OK();
}

template <PrimitiveType PT>
Status FixedSizeJoinBuildFunc<PT>::prepare(RuntimeState* state, JoinHashTableItems* table_items,
                                           HashTableProbeState* probe_state) {
    table_items->build_key_column = ColumnType::create(table_items->row_count + 1);
    probe_state->buckets.resize(config::vector_chunk_size);
    probe_state->is_nulls.resize(config::vector_chunk_size);
    return Status::OK();
}

template <PrimitiveType PT>
Status FixedSizeJoinBuildFunc<PT>::construct_hash_table(JoinHashTableItems* table_items,
                                                        HashTableProbeState* probe_state) {
    uint32_t row_count = table_items->row_count;

    // prepare columns
    Columns data_columns;
    NullColumns null_columns;
    for (size_t i = 0; i < table_items->key_columns.size(); i++) {
        if (table_items->join_keys[i].is_null_safe_equal) {
            data_columns.emplace_back(table_items->key_columns[i]);
        } else if (table_items->key_columns[i]->is_nullable()) {
            auto* nullable_column =
                    ColumnHelper::as_raw_column<NullableColumn>(table_items->key_columns[i]);
            data_columns.emplace_back(nullable_column->data_column());
            if (table_items->key_columns[i]->has_null()) {
                null_columns.emplace_back(nullable_column->null_column());
            }
        } else {
            data_columns.emplace_back(table_items->key_columns[i]);
        }
    }

    // serialize and build hash table
    uint32_t quo = row_count / config::vector_chunk_size;
    uint32_t rem = row_count % config::vector_chunk_size;

    if (!null_columns.empty()) {
        for (size_t i = 0; i < quo; i++) {
            _build_nullable_columns(table_items, probe_state, data_columns, null_columns,
                                    1 + config::vector_chunk_size * i, config::vector_chunk_size);
        }
        _build_nullable_columns(table_items, probe_state, data_columns, null_columns,
                                1 + config::vector_chunk_size * quo, rem);
    } else {
        for (size_t i = 0; i < quo; i++) {
            _build_columns(table_items, probe_state, data_columns,
                           1 + config::vector_chunk_size * i, config::vector_chunk_size);
        }
        _build_columns(table_items, probe_state, data_columns, 1 + config::vector_chunk_size * quo,
                       rem);
    }
    return Status::OK();
}

template <PrimitiveType PT>
void FixedSizeJoinBuildFunc<PT>::_build_columns(JoinHashTableItems* table_items,
                                                HashTableProbeState* probe_state,
                                                const Columns& data_columns, uint32_t start,
                                                uint32_t count) {
    JoinHashMapHelper::serialize_fixed_size_key_column<PT>(
            data_columns, table_items->build_key_column.get(), start, count);

    const auto& data = get_key_data(*table_items);
    JoinHashMapHelper::calc_bucket_nums<CppType>(data, table_items->bucket_size,
                                                 &probe_state->buckets, start, count);

    for (uint32_t i = 0; i < count; i++) {
        table_items->next[start + i] = table_items->first[probe_state->buckets[i]];
        table_items->first[probe_state->buckets[i]] = start + i;
    }
}

template <PrimitiveType PT>
void FixedSizeJoinBuildFunc<PT>::_build_nullable_columns(JoinHashTableItems* table_items,
                                                         HashTableProbeState* probe_state,
                                                         const Columns& data_columns,
                                                         const NullColumns& null_columns,
                                                         uint32_t start, uint32_t count) {
    for (uint32_t i = 0; i < count; i++) {
        probe_state->is_nulls[i] = null_columns[0]->get_data()[start + i];
    }
    for (uint32_t i = 1; i < null_columns.size(); i++) {
        for (uint32_t j = 0; j < count; j++) {
            probe_state->is_nulls[j] |= null_columns[i]->get_data()[start + j];
        }
    }

    JoinHashMapHelper::serialize_fixed_size_key_column<PT>(
            data_columns, table_items->build_key_column.get(), start, count);
    const auto& data = get_key_data(*table_items);
    JoinHashMapHelper::calc_bucket_nums<CppType>(data, table_items->bucket_size,
                                                 &probe_state->buckets, start, count);

    for (size_t i = 0; i < count; i++) {
        if (probe_state->is_nulls[i] == 0) {
            table_items->next[start + i] = table_items->first[probe_state->buckets[i]];
            table_items->first[probe_state->buckets[i]] = start + i;
        }
    }
}

template <PrimitiveType PT>
Status JoinProbeFunc<PT>::lookup_init(const JoinHashTableItems& table_items,
                                      HashTableProbeState* probe_state) {
    size_t probe_row_count = probe_state->probe_row_count;
    auto& data = get_key_data(*probe_state);
    JoinHashMapHelper::calc_bucket_nums<CppType>(data, table_items.bucket_size,
                                                 &probe_state->buckets, 0, data.size());

    if ((*probe_state->key_columns)[0]->is_nullable()) {
        auto* nullable_column =
                ColumnHelper::as_raw_column<NullableColumn>((*probe_state->key_columns)[0]);

        if (nullable_column->has_null()) {
            auto& null_array = nullable_column->null_column()->get_data();
            for (size_t i = 0; i < probe_row_count; i++) {
                if (null_array[i] == 0) {
                    probe_state->next[i] = table_items.first[probe_state->buckets[i]];
                } else {
                    probe_state->next[i] = 0;
                }
            }
            probe_state->null_array = &nullable_column->null_column()->get_data();
        } else {
            for (size_t i = 0; i < probe_row_count; i++) {
                probe_state->next[i] = table_items.first[probe_state->buckets[i]];
            }
            probe_state->null_array = nullptr;
        }
        return Status::OK();
    }

    for (size_t i = 0; i < probe_row_count; i++) {
        probe_state->next[i] = table_items.first[probe_state->buckets[i]];
    }
    probe_state->null_array = nullptr;
    return Status::OK();
}

template <PrimitiveType PT>
const Buffer<typename JoinProbeFunc<PT>::CppType>& JoinProbeFunc<PT>::get_key_data(
        const HashTableProbeState& probe_state) {
    if ((*probe_state.key_columns)[0]->is_nullable()) {
        auto* nullable_column =
                ColumnHelper::as_raw_column<NullableColumn>((*probe_state.key_columns)[0]);
        return ColumnHelper::as_raw_column<ColumnType>(nullable_column->data_column())->get_data();
    }

    return ColumnHelper::as_raw_column<ColumnType>((*probe_state.key_columns)[0])->get_data();
}

template <PrimitiveType PT>
Status FixedSizeJoinProbeFunc<PT>::lookup_init(const JoinHashTableItems& table_items,
                                               HashTableProbeState* probe_state) {
    // prepare columns
    Columns data_columns;
    NullColumns null_columns;

    for (size_t i = 0; i < probe_state->key_columns->size(); i++) {
        if (table_items.join_keys[i].is_null_safe_equal) {
            if ((*probe_state->key_columns)[i]->is_nullable()) {
                data_columns.emplace_back((*probe_state->key_columns)[i]);
            } else {
                auto tmp_column =
                        NullableColumn::create((*probe_state->key_columns)[i],
                                               NullColumn::create(probe_state->probe_row_count, 0));
                data_columns.emplace_back(tmp_column);
            }
        } else if ((*probe_state->key_columns)[i]->is_nullable()) {
            auto* nullable_column =
                    ColumnHelper::as_raw_column<NullableColumn>((*probe_state->key_columns)[i]);
            data_columns.emplace_back(nullable_column->data_column());
            if ((*probe_state->key_columns)[i]->has_null()) {
                null_columns.emplace_back(nullable_column->null_column());
            }
        } else {
            data_columns.emplace_back((*probe_state->key_columns)[i]);
        }
    }

    // serialize and init search
    if (!null_columns.empty()) {
        _probe_nullable_column(table_items, probe_state, data_columns, null_columns);
    } else {
        _probe_column(table_items, probe_state, data_columns);
    }
    return Status::OK();
}

template <PrimitiveType PT>
void FixedSizeJoinProbeFunc<PT>::_probe_column(const JoinHashTableItems& table_items,
                                               HashTableProbeState* probe_state,
                                               const Columns& data_columns) {
    uint32_t row_count = probe_state->probe_row_count;

    JoinHashMapHelper::serialize_fixed_size_key_column<PT>(
            data_columns, probe_state->probe_key_column.get(), 0, row_count);
    const auto& data = get_key_data(*probe_state);
    JoinHashMapHelper::calc_bucket_nums<CppType>(data, table_items.bucket_size,
                                                 &probe_state->buckets, 0, row_count);

    for (uint32_t i = 0; i < row_count; i++) {
        probe_state->next[i] = table_items.first[probe_state->buckets[i]];
    }
}

template <PrimitiveType PT>
void FixedSizeJoinProbeFunc<PT>::_probe_nullable_column(const JoinHashTableItems& table_items,
                                                        HashTableProbeState* probe_state,
                                                        const Columns& data_columns,
                                                        const NullColumns& null_columns) {
    uint32_t row_count = probe_state->probe_row_count;

    for (uint32_t i = 0; i < row_count; i++) {
        probe_state->is_nulls[i] = null_columns[0]->get_data()[i];
    }
    for (uint32_t i = 1; i < null_columns.size(); i++) {
        for (uint32_t j = 0; j < row_count; j++) {
            probe_state->is_nulls[j] |= null_columns[i]->get_data()[j];
        }
    }

    JoinHashMapHelper::serialize_fixed_size_key_column<PT>(
            data_columns, probe_state->probe_key_column.get(), 0, row_count);
    const auto& data = get_key_data(*probe_state);
    JoinHashMapHelper::calc_bucket_nums<CppType>(data, table_items.bucket_size,
                                                 &probe_state->buckets, 0, row_count);

    for (uint32_t i = 0; i < row_count; i++) {
        if (probe_state->is_nulls[i] == 0) {
            probe_state->next[i] = table_items.first[probe_state->buckets[i]];
        } else {
            probe_state->next[i] = 0;
        }
    }
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::build(RuntimeState* state) {
    // prepare
    RETURN_IF_ERROR(BuildFunc().prepare(state, _table_items, _probe_state));

    // construct hash table
    RETURN_IF_ERROR(BuildFunc().construct_hash_table(_table_items, _probe_state));

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::probe(const Columns& key_columns,
                                                    ChunkPtr* probe_chunk, ChunkPtr* chunk,
                                                    bool* has_remain) {
    _probe_state->key_columns = &key_columns;
    {
        SCOPED_TIMER(_table_items->search_ht_timer);
        RETURN_IF_ERROR(_search_ht(probe_chunk));
        if (_probe_state->count <= 0) {
            *has_remain = false;
            return Status::OK();
        }

        *has_remain = _probe_state->has_remain;
    }

    if (_table_items->join_type == TJoinOp::RIGHT_SEMI_JOIN ||
        _table_items->join_type == TJoinOp::RIGHT_ANTI_JOIN) {
        // right semi join without other join conjunct
        // right anti join without other join conjunct
        // don't need output the real probe column
        {
            // output default values for probe-columns as placeholder.
            SCOPED_TIMER(_table_items->output_probe_column_timer);
            if (!_table_items->with_other_conjunct) {
                RETURN_IF_ERROR(_probe_null_output(chunk, _probe_state->count));
            } else {
                RETURN_IF_ERROR(_probe_output(probe_chunk, chunk));
            }
        }
        {
            SCOPED_TIMER(_table_items->output_build_column_timer);
            RETURN_IF_ERROR(_build_output(chunk));
        }
        {
            SCOPED_TIMER(_table_items->output_tuple_column_timer);
            _build_tuple_output(chunk);
        }
    } else if (_table_items->join_type == TJoinOp::LEFT_SEMI_JOIN ||
               _table_items->join_type == TJoinOp::LEFT_ANTI_JOIN ||
               _table_items->join_type == TJoinOp::NULL_AWARE_LEFT_ANTI_JOIN) {
        // left semi join without other join conjunct
        // anti anti join without other join conjunct
        // don't need output the real build column
        {
            SCOPED_TIMER(_table_items->output_probe_column_timer);
            RETURN_IF_ERROR(_probe_output(probe_chunk, chunk));
        }
        {
            // output default values for build-columns as placeholder.
            SCOPED_TIMER(_table_items->output_build_column_timer);
            if (!_table_items->with_other_conjunct) {
                RETURN_IF_ERROR(_build_default_output(chunk, _probe_state->count));
            } else {
                RETURN_IF_ERROR(_build_output(chunk));
            }
        }
        {
            SCOPED_TIMER(_table_items->output_tuple_column_timer);
            _probe_tuple_output(probe_chunk, chunk);
        }
    } else {
        {
            SCOPED_TIMER(_table_items->output_probe_column_timer);
            RETURN_IF_ERROR(_probe_output(probe_chunk, chunk));
        }
        {
            SCOPED_TIMER(_table_items->output_build_column_timer);
            RETURN_IF_ERROR(_build_output(chunk));
        }
        {
            SCOPED_TIMER(_table_items->output_tuple_column_timer);
            _probe_tuple_output(probe_chunk, chunk);
            _build_tuple_output(chunk);
        }
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::probe_remain(ChunkPtr* chunk, bool* has_remain) {
    _search_ht_remain();
    if (_probe_state->count <= 0) {
        *has_remain = false;
        return Status::OK();
    }
    *has_remain = _probe_state->has_remain;

    if (_table_items->join_type == TJoinOp::RIGHT_ANTI_JOIN ||
        _table_items->join_type == TJoinOp::RIGHT_SEMI_JOIN) {
        // right anti/semi join without other conjunct output default value of probe-columns as placeholder.
        RETURN_IF_ERROR(_probe_null_output(chunk, _probe_state->count));
        RETURN_IF_ERROR(_build_output(chunk));

        _build_tuple_output(chunk);
    } else {
        // RIGHT_OUTER_JOIN || FULL_OUTER_JOIN
        RETURN_IF_ERROR(_probe_null_output(chunk, _probe_state->count));
        RETURN_IF_ERROR(_build_output(chunk));

        for (int tuple_id : _table_items->output_probe_tuple_ids) {
            ColumnPtr column = BooleanColumn::create(_probe_state->count, 0);
            (*chunk)->append_tuple_column(column, tuple_id);
        }
        _build_tuple_output(chunk);
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_output(ChunkPtr* probe_chunk,
                                                            ChunkPtr* chunk) {
    bool to_nullable = _table_items->left_to_nullable;

    for (size_t i = 0; i < _table_items->probe_column_count; i++) {
        SlotDescriptor* slot = _table_items->probe_slots[i];
        auto& column = (*probe_chunk)->get_column_by_slot_id(slot->id());
        if (!column->is_nullable()) {
            _copy_probe_column(&column, chunk, slot, to_nullable);
        } else {
            _copy_probe_nullable_column(&column, chunk, slot);
        }
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_tuple_output(ChunkPtr* probe_chunk,
                                                                ChunkPtr* chunk) {
    bool to_nullable = _table_items->left_to_nullable;

    for (int tuple_id : _table_items->output_probe_tuple_ids) {
        //TODO: only calc one time
        if ((*probe_chunk)->is_tuple_exist(tuple_id)) {
            ColumnPtr& src_column = (*probe_chunk)->get_tuple_column_by_id(tuple_id);
            if (_probe_state->match_flag == JoinMatchFlag::ALL_MATCH_ONE) {
                (*chunk)->append_tuple_column(src_column, tuple_id);
            } else if (_probe_state->match_flag == JoinMatchFlag::MOST_MATCH_ONE) {
                src_column->filter(_probe_state->probe_match_filter, _probe_state->probe_row_count);
                (*chunk)->append_tuple_column(src_column, tuple_id);
            } else {
                ColumnPtr dest_column = BooleanColumn::create(_probe_state->count);
                auto& src_data = ColumnHelper::as_raw_column<BooleanColumn>(src_column)->get_data();
                auto& dest_data =
                        ColumnHelper::as_raw_column<BooleanColumn>(dest_column)->get_data();

                size_t end = _probe_state->count;
                for (size_t i = 0; i < end; i++) {
                    dest_data[i] = src_data[_probe_state->probe_index[i]];
                }
                (*chunk)->append_tuple_column(dest_column, tuple_id);
            }
        } else {
            if (to_nullable) {
                ColumnPtr dest_column = BooleanColumn::create(_probe_state->count, 1);
                (*chunk)->append_tuple_column(dest_column, tuple_id);
            }
        }
    }
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_null_output(ChunkPtr* chunk, size_t count) {
    for (size_t i = 0; i < _table_items->probe_column_count; i++) {
        SlotDescriptor* slot = _table_items->probe_slots[i];
        ColumnPtr column = ColumnHelper::create_column(slot->type(), true);
        column->append_nulls(count);
        (*chunk)->append_column(std::move(column), slot->id());
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::_build_output(ChunkPtr* chunk) {
    bool to_nullable = _table_items->right_to_nullable;
    for (size_t i = 0; i < _table_items->build_column_count; i++) {
        SlotDescriptor* slot = _table_items->build_slots[i];
        ColumnPtr& column = _table_items->build_chunk->columns()[i];
        if (!column->is_nullable()) {
            _copy_build_column(column, chunk, slot, to_nullable);
        } else {
            _copy_build_nullable_column(column, chunk, slot);
        }
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_build_tuple_output(ChunkPtr* chunk) {
    // TODO: performance
    // The function will be deleted when IfTupleIsNull is removed
    bool to_nullable = _table_items->right_to_nullable;

    for (int tuple_id : _table_items->output_build_tuple_ids) {
        if (_table_items->build_chunk->is_tuple_exist(tuple_id)) {
            // tuple column exist
            ColumnPtr& src_column = _table_items->build_chunk->get_tuple_column_by_id(tuple_id);
            auto& src_data = ColumnHelper::as_raw_column<BooleanColumn>(src_column)->get_data();
            ColumnPtr dest_column = BooleanColumn::create(_probe_state->count);
            auto& dest_data = ColumnHelper::as_raw_column<BooleanColumn>(dest_column)->get_data();
            size_t row_count = _probe_state->count;

            if (!_probe_state->has_null_build_tuple) {
                // all match
                for (size_t i = 0; i < row_count; i++) {
                    dest_data[i] = src_data[_probe_state->build_index[i]];
                }
            } else {
                for (size_t i = 0; i < row_count; i++) {
                    dest_data[i] = (_probe_state->build_index[i] != 0) &
                                   src_data[_probe_state->build_index[i]];
                }
            }
            (*chunk)->append_tuple_column(dest_column, tuple_id);
        } else {
            // tuple column not exist
            if (to_nullable) {
                if (!_probe_state->has_null_build_tuple) {
                    // all match
                    ColumnPtr dest_column = BooleanColumn::create(_probe_state->count, 1);
                    (*chunk)->append_tuple_column(dest_column, tuple_id);
                } else {
                    ColumnPtr dest_column = BooleanColumn::create(_probe_state->count);
                    auto& dest_data =
                            ColumnHelper::as_raw_column<BooleanColumn>(dest_column)->get_data();

                    size_t end = _probe_state->count;
                    for (size_t i = 0; i < end; i++) {
                        dest_data[i] = _probe_state->build_index[i] != 0;
                    }
                    (*chunk)->append_tuple_column(dest_column, tuple_id);
                }
            }
        }
    }
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::_build_default_output(ChunkPtr* chunk, size_t count) {
    for (size_t i = 0; i < _table_items->build_column_count; i++) {
        SlotDescriptor* slot = _table_items->build_slots[i];
        ColumnPtr column = ColumnHelper::create_column(slot->type(), true);
        column->append_nulls(count);
        (*chunk)->append_column(std::move(column), slot->id());
    }

    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_copy_probe_column(ColumnPtr* src_column,
                                                               ChunkPtr* chunk,
                                                               const SlotDescriptor* slot,
                                                               bool to_nullable) {
    if (_probe_state->match_flag == JoinMatchFlag::ALL_MATCH_ONE) {
        if (to_nullable) {
            ColumnPtr dest_column =
                    NullableColumn::create(*src_column, NullColumn::create((*src_column)->size()));
            (*chunk)->append_column(std::move(dest_column), slot->id());
        } else {
            (*chunk)->append_column(*src_column, slot->id());
        }
    } else if (_probe_state->match_flag == JoinMatchFlag::MOST_MATCH_ONE) {
        if (to_nullable) {
            (*src_column)->filter(_probe_state->probe_match_filter, _probe_state->probe_row_count);
            ColumnPtr dest_column =
                    NullableColumn::create(*src_column, NullColumn::create((*src_column)->size()));
            (*chunk)->append_column(std::move(dest_column), slot->id());
        } else {
            (*src_column)->filter(_probe_state->probe_match_filter, _probe_state->probe_row_count);
            (*chunk)->append_column(*src_column, slot->id());
        }
    } else {
        ColumnPtr dest_column = ColumnHelper::create_column(slot->type(), to_nullable);
        dest_column->append_selective(**src_column, _probe_state->probe_index.data(), 0,
                                      _probe_state->count);
        (*chunk)->append_column(std::move(dest_column), slot->id());
    }
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_copy_probe_nullable_column(
        ColumnPtr* src_column, ChunkPtr* chunk, const SlotDescriptor* slot) {
    if (_probe_state->match_flag == JoinMatchFlag::ALL_MATCH_ONE) {
        (*chunk)->append_column(*src_column, slot->id());
    } else if (_probe_state->match_flag == JoinMatchFlag::MOST_MATCH_ONE) {
        (*src_column)->filter(_probe_state->probe_match_filter, _probe_state->probe_row_count);
        (*chunk)->append_column(*src_column, slot->id());
    } else {
        ColumnPtr dest_column = ColumnHelper::create_column(slot->type(), true);
        dest_column->append_selective(**src_column, _probe_state->probe_index.data(), 0,
                                      _probe_state->count);
        (*chunk)->append_column(std::move(dest_column), slot->id());
    }
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_copy_build_column(const ColumnPtr& src_column,
                                                               ChunkPtr* chunk,
                                                               const SlotDescriptor* slot,
                                                               bool to_nullable) {
    ColumnPtr dest_column = ColumnHelper::create_column(slot->type(), to_nullable);

    if (to_nullable) {
        dest_column->append_selective(*src_column, _probe_state->build_index.data(), 0,
                                      _probe_state->count);

        // When left outer join is executed,
        // build_index[i] Equal to 0 means it is not found in the hash table,
        // but append_selective() has set item of NullColumn to not null
        // so NullColumn needs to be set back to null
        auto* null_column = ColumnHelper::as_raw_column<NullableColumn>(dest_column);
        size_t end = _probe_state->count;
        for (size_t i = 0; i < end; i++) {
            if (_probe_state->build_index[i] == 0) {
                null_column->set_null(i);
            }
        }
    } else {
        dest_column->append_selective(*src_column, _probe_state->build_index.data(), 0,
                                      _probe_state->count);
    }

    (*chunk)->append_column(std::move(dest_column), slot->id());
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_copy_build_nullable_column(
        const ColumnPtr& src_column, ChunkPtr* chunk, const SlotDescriptor* slot) {
    ColumnPtr dest_column = ColumnHelper::create_column(slot->type(), true);

    dest_column->append_selective(*src_column, _probe_state->build_index.data(), 0,
                                  _probe_state->count);

    // When left outer join is executed,
    // build_index[i] Equal to 0 means it is not found in the hash table,
    // but append_selective() has set item of NullColumn to not null
    // so NullColumn needs to be set back to null
    auto* null_column = ColumnHelper::as_raw_column<NullableColumn>(dest_column);
    size_t end = _probe_state->count;
    for (size_t i = 0; i < end; i++) {
        if (_probe_state->build_index[i] == 0) {
            null_column->set_null(i);
        }
    }

    (*chunk)->append_column(std::move(dest_column), slot->id());
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
Status JoinHashMap<PT, BuildFunc, ProbeFunc>::_search_ht(ChunkPtr* probe_chunk) {
    if (!_probe_state->has_remain) {
        _probe_state->probe_row_count = (*probe_chunk)->num_rows();
        ProbeFunc().prepare(_table_items, _probe_state);
        RETURN_IF_ERROR(ProbeFunc().lookup_init(*_table_items, _probe_state));

        auto& build_data = BuildFunc().get_key_data(*_table_items);
        auto& probe_data = ProbeFunc().get_key_data(*_probe_state);
        _search_ht_impl<true>(build_data, probe_data);
    } else {
        auto& build_data = BuildFunc().get_key_data(*_table_items);
        auto& probe_data = ProbeFunc().get_key_data(*_probe_state);
        _search_ht_impl<false>(build_data, probe_data);
    }
    return Status::OK();
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_search_ht_remain() {
    if (!_probe_state->has_remain) {
        size_t zero_count = SIMD::count_zero(_probe_state->build_match_index);
        if (zero_count <= 0) {
            _probe_state->count = 0;
            _probe_state->has_remain = false;
            return;
        }
        _probe_state->cur_probe_index = 0;
    }

    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;
    for (; i < _probe_state->build_match_index.size(); i++) {
        if (_probe_state->build_match_index[i] == 0) {
            _probe_state->build_index[match_count] = i;
            _probe_state->probe_index[match_count] = 0;
            match_count++;

            if (match_count >= config::vector_chunk_size) {
                i++;
                _probe_state->cur_probe_index = i;
                _probe_state->has_remain = i < _probe_state->build_match_index.size();
                _probe_state->count = match_count;
                return;
            }
        }
    }

    _probe_state->cur_probe_index = i;
    _probe_state->has_remain = false;
    _probe_state->count = match_count;
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_search_ht_impl(const Buffer<CppType>& build_data,
                                                            const Buffer<CppType>& data) {
    if (!_table_items->with_other_conjunct) {
        switch (_table_items->join_type) {
        case TJoinOp::LEFT_OUTER_JOIN:
            _probe_from_ht_for_left_outer_join<first_probe>(build_data, data);
            break;
        case TJoinOp::LEFT_SEMI_JOIN:
            _probe_from_ht_for_left_semi_join<first_probe>(build_data, data);
            break;
        case TJoinOp::LEFT_ANTI_JOIN:
        case TJoinOp::NULL_AWARE_LEFT_ANTI_JOIN:
            _probe_from_ht_for_left_anti_join<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_OUTER_JOIN:
            _probe_from_ht_for_right_outer_join<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_SEMI_JOIN:
            _probe_from_ht_for_right_semi_join<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_ANTI_JOIN:
            _probe_from_ht_for_right_anti_join<first_probe>(build_data, data);
            break;
        case TJoinOp::FULL_OUTER_JOIN:
            _probe_from_ht_for_full_outer_join<first_probe>(build_data, data);
            break;
        default:
            _probe_from_ht<first_probe>(build_data, data);
            break;
        }
    } else {
        switch (_table_items->join_type) {
        case TJoinOp::LEFT_OUTER_JOIN:
            _probe_from_ht_for_left_outer_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::LEFT_SEMI_JOIN:
            _probe_from_ht_for_left_semi_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::LEFT_ANTI_JOIN:
        case TJoinOp::NULL_AWARE_LEFT_ANTI_JOIN:
            _probe_from_ht_for_left_anti_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_OUTER_JOIN:
            _probe_from_ht_for_right_outer_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_SEMI_JOIN:
            _probe_from_ht_for_right_semi_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::RIGHT_ANTI_JOIN:
            _probe_from_ht_for_right_anti_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        case TJoinOp::FULL_OUTER_JOIN:
            _probe_from_ht_for_full_outer_join_with_other_conjunct<first_probe>(build_data, data);
            break;
        default:
            // can't reach here
            _probe_from_ht<first_probe>(build_data, data);
            break;
        }
    }
}

#define CHECK_MATCH()                                                                              \
    if (match_count > 0 && !one_to_many) {                                                         \
        size_t zero_count =                                                                        \
                SIMD::count_zero(_probe_state->probe_match_filter, _probe_state->probe_row_count); \
        if (zero_count == 0) {                                                                     \
            _probe_state->match_flag = JoinMatchFlag::ALL_MATCH_ONE;                               \
        } else if (zero_count < _probe_state->probe_row_count - zero_count) {                      \
            _probe_state->match_flag = JoinMatchFlag::MOST_MATCH_ONE;                              \
        }                                                                                          \
    }

#define CHECK_ALL_MATCH()                                        \
    if (match_count > 0 && !one_to_many) {                       \
        _probe_state->match_flag = JoinMatchFlag::ALL_MATCH_ONE; \
    }

#define RETURN_IF_CHUNK_FULL()                                   \
    if (match_count > config::vector_chunk_size) {               \
        _probe_state->next[i] = _table_items->next[build_index]; \
        _probe_state->cur_probe_index = i;                       \
        _probe_state->has_remain = true;                         \
        _probe_state->count = config::vector_chunk_size;         \
        return;                                                  \
    }

// When a probe row corresponds to multiple Build rows,
// a Probe Chunk may generate multiple ResultChunks,
// so each probe will have search one more row to determine whether it has reached the boundary,
// so the next probe will start from the last recorded position
#define PROCESS_PROBE_STAGE_FOR_RIGHT_JOIN_WITH_OTHER_CONJUNCT()                             \
    if constexpr (!first_probe) {                                                            \
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size]; \
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size]; \
        match_count = 1;                                                                     \
        if (_probe_state->next[i] == 0) {                                                    \
            i++;                                                                             \
        }                                                                                    \
    }

#define PROBE_OVER()                   \
    _probe_state->has_remain = false;  \
    _probe_state->cur_probe_index = 0; \
    _probe_state->count = match_count; \
    _probe_state->cur_row_match_count = 0;

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht(const Buffer<CppType>& build_data,
                                                           const Buffer<CppType>& probe_data) {
    _probe_state->match_flag = JoinMatchFlag::NORMAL;
    size_t match_count = 0;
    bool one_to_many = false;
    size_t i = _probe_state->cur_probe_index;

    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        if constexpr (first_probe) {
            _probe_state->probe_match_filter[i] = 0;
        }
        size_t build_index = _probe_state->next[i];
        if (build_index != 0) {
            do {
                if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                    _probe_state->probe_index[match_count] = i;
                    _probe_state->build_index[match_count] = build_index;
                    match_count++;

                    if constexpr (first_probe) {
                        _probe_state->cur_row_match_count++;
                        _probe_state->probe_match_filter[i] = 1;
                    }
                    RETURN_IF_CHUNK_FULL()
                }
                build_index = _table_items->next[build_index];
            } while (build_index != 0);

            if constexpr (first_probe) {
                if (_probe_state->cur_row_match_count > 1) {
                    one_to_many = true;
                }
            }
        }

        if constexpr (first_probe) {
            _probe_state->cur_row_match_count = 0;
        }
    }

    if constexpr (first_probe) {
        CHECK_MATCH()
    }
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_outer_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    _probe_state->match_flag = JoinMatchFlag::NORMAL;
    size_t match_count = 0;
    bool one_to_many = false;
    size_t i = _probe_state->cur_probe_index;

    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;
            _probe_state->has_null_build_tuple = true;

            RETURN_IF_CHUNK_FULL()
        } else {
            while (build_index != 0) {
                if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                    _probe_state->probe_index[match_count] = i;
                    _probe_state->build_index[match_count] = build_index;
                    match_count++;
                    _probe_state->cur_row_match_count++;

                    RETURN_IF_CHUNK_FULL()
                }
                build_index = _table_items->next[build_index];
            }
            if (_probe_state->cur_row_match_count <= 0) {
                // one key of left table match none key of right table
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = 0;
                match_count++;
                _probe_state->has_null_build_tuple = true;

                RETURN_IF_CHUNK_FULL()
            } else if (_probe_state->cur_row_match_count > 1) {
                // one key of left table match multi key of right table
                if constexpr (first_probe) {
                    one_to_many = true;
                }
            }
        }
        _probe_state->cur_row_match_count = 0;
    }

    if constexpr (first_probe) {
        CHECK_ALL_MATCH()
    }
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_semi_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t probe_row_count = _probe_state->probe_row_count;
    for (size_t i = 0; i < probe_row_count; i++) {
        size_t index = _probe_state->next[i];
        if (index == 0) {
            continue;
        }

        while (index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                match_count++;
                break;
            }
            index = _table_items->next[index];
        }
    }

    PROBE_OVER()
    _probe_state->has_null_build_tuple = true;
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_anti_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;

    size_t probe_row_count = _probe_state->probe_row_count;
    if (_table_items->join_type == TJoinOp::NULL_AWARE_LEFT_ANTI_JOIN &&
        _probe_state->null_array != nullptr && _table_items->row_count != 0) {
        // process left anti join from not in
        for (size_t i = 0; i < probe_row_count; i++) {
            size_t index = _probe_state->next[i];
            if ((*_probe_state->null_array)[i] == 1) {
                continue;
            }

            if (index == 0) {
                _probe_state->probe_index[match_count] = i;
                match_count++;
                continue;
            }

            bool found = false;
            while (index != 0) {
                if (JoinKeyEqual<CppType>()(build_data[index], probe_data[i])) {
                    found = true;
                    break;
                }
                index = _table_items->next[index];
            }
            if (!found) {
                _probe_state->probe_index[match_count] = i;
                match_count++;
            }
        }
    } else {
        for (size_t i = 0; i < probe_row_count; i++) {
            size_t index = _probe_state->next[i];
            if (index == 0) {
                _probe_state->probe_index[match_count] = i;
                match_count++;
                continue;
            }
            bool found = false;
            while (index != 0) {
                if (JoinKeyEqual<CppType>()(build_data[index], probe_data[i])) {
                    found = true;
                    break;
                }
                index = _table_items->next[index];
            }
            if (!found) {
                _probe_state->probe_index[match_count] = i;
                match_count++;
            }
        }
    }

    PROBE_OVER()
    _probe_state->has_null_build_tuple = true;
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_outer_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                _probe_state->build_match_index[build_index] = 1;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
    }

    // TODO: all match optimized
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_semi_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                if (_probe_state->build_match_index[build_index] == 0) {
                    _probe_state->probe_index[match_count] = i;
                    _probe_state->build_index[match_count] = build_index;
                    _probe_state->build_match_index[build_index] = 1;
                    match_count++;

                    RETURN_IF_CHUNK_FULL()
                }
            }
            build_index = _table_items->next[build_index];
        }
    }

    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_anti_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t probe_row_count = _probe_state->probe_row_count;
    for (size_t i = 0; i < probe_row_count; i++) {
        size_t index = _probe_state->next[i];
        if (index == 0) {
            continue;
        }

        while (index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[index], probe_data[i])) {
                _probe_state->build_match_index[index] = 1;
            }
            index = _table_items->next[index];
        }
    }
    _probe_state->count = 0;
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_full_outer_join(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    } else {
        _probe_state->cur_row_match_count = 0;
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
        } else {
            while (build_index != 0) {
                if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                    _probe_state->probe_index[match_count] = i;
                    _probe_state->build_index[match_count] = build_index;
                    _probe_state->build_match_index[build_index] = 1;
                    _probe_state->cur_row_match_count++;
                    match_count++;

                    RETURN_IF_CHUNK_FULL()
                }
                build_index = _table_items->next[build_index];
            }
            if (_probe_state->cur_row_match_count <= 0) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = 0;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
        }
        _probe_state->cur_row_match_count = 0;
    }

    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_outer_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;

    size_t i = _probe_state->cur_probe_index;
    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    } else {
        _probe_state->cur_row_match_count = 0;
        for (size_t j = 0; j < config::vector_chunk_size; j++) {
            _probe_state->probe_match_index[j] = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                _probe_state->probe_match_index[i]++;
                match_count++;
                _probe_state->cur_row_match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
        if (_probe_state->cur_row_match_count <= 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
        }
        _probe_state->cur_row_match_count = 0;
    }

    _probe_state->has_null_build_tuple = true;
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_semi_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;

    size_t i = _probe_state->cur_probe_index;
    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    } else {
        for (size_t j = 0; j < config::vector_chunk_size; j++) {
            _probe_state->probe_match_index[j] = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
    }

    _probe_state->has_null_build_tuple = true;
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_left_anti_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;

    size_t i = _probe_state->cur_probe_index;
    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    } else {
        _probe_state->cur_row_match_count = 0;
        for (size_t j = 0; j < config::vector_chunk_size; j++) {
            _probe_state->probe_match_index[j] = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                _probe_state->probe_match_index[i]++;
                match_count++;
                _probe_state->cur_row_match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
        if (_probe_state->cur_row_match_count <= 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
        }
        _probe_state->cur_row_match_count = 0;
    }

    _probe_state->has_null_build_tuple = true;
    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_outer_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    PROCESS_PROBE_STAGE_FOR_RIGHT_JOIN_WITH_OTHER_CONJUNCT()

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
    }

    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_semi_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    PROCESS_PROBE_STAGE_FOR_RIGHT_JOIN_WITH_OTHER_CONJUNCT()

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
    }

    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_right_anti_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;
    size_t i = _probe_state->cur_probe_index;

    PROCESS_PROBE_STAGE_FOR_RIGHT_JOIN_WITH_OTHER_CONJUNCT()

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            continue;
        }

        while (build_index != 0) {
            if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = build_index;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
            build_index = _table_items->next[build_index];
        }
    }

    PROBE_OVER()
}

template <PrimitiveType PT, class BuildFunc, class ProbeFunc>
template <bool first_probe>
void JoinHashMap<PT, BuildFunc, ProbeFunc>::_probe_from_ht_for_full_outer_join_with_other_conjunct(
        const Buffer<CppType>& build_data, const Buffer<CppType>& probe_data) {
    size_t match_count = 0;

    size_t i = _probe_state->cur_probe_index;
    if constexpr (!first_probe) {
        _probe_state->probe_index[0] = _probe_state->probe_index[config::vector_chunk_size];
        _probe_state->build_index[0] = _probe_state->build_index[config::vector_chunk_size];
        match_count = 1;
        if (_probe_state->next[i] == 0) {
            i++;
            _probe_state->cur_row_match_count = 0;
        }
    } else {
        _probe_state->cur_row_match_count = 0;
        for (size_t j = 0; j < config::vector_chunk_size; j++) {
            _probe_state->probe_match_index[j] = 0;
        }
    }

    size_t probe_row_count = _probe_state->probe_row_count;
    for (; i < probe_row_count; i++) {
        size_t build_index = _probe_state->next[i];
        if (build_index == 0) {
            _probe_state->probe_index[match_count] = i;
            _probe_state->build_index[match_count] = 0;
            match_count++;

            RETURN_IF_CHUNK_FULL()
        } else {
            while (build_index != 0) {
                if (JoinKeyEqual<CppType>()(build_data[build_index], probe_data[i])) {
                    _probe_state->probe_index[match_count] = i;
                    _probe_state->build_index[match_count] = build_index;
                    _probe_state->probe_match_index[i]++;
                    _probe_state->cur_row_match_count++;
                    match_count++;

                    RETURN_IF_CHUNK_FULL()
                }
                build_index = _table_items->next[build_index];
            }
            if (_probe_state->cur_row_match_count <= 0) {
                _probe_state->probe_index[match_count] = i;
                _probe_state->build_index[match_count] = 0;
                match_count++;

                RETURN_IF_CHUNK_FULL()
            }
        }
        _probe_state->cur_row_match_count = 0;
    }

    _probe_state->has_null_build_tuple = true;

    PROBE_OVER()
}

} // namespace starrocks::vectorized
